#ifndef __MATRIX_H_INCLUDED__
#define __MATRIX_H_INCLUDED__

///////////////////////////////////////////////////////////////////////////////
typedef struct _matrix_internal
{
 union {
        struct  {				//  v--- unused
                  float xx, xy, xz, xw;
                  float yx, yy, yz, yw;
                  float zx, zy, zz, zw;
                  float wx, wy, wz, ww; // <-- translation
                };
        struct  {
                  float  _00, _01, _02, _03;
                  float  _10, _11, _12, _13;
                  float  _20, _21, _22, _23;
                  float  _30, _31, _32, _33;
                };
        struct  {
                  float v[16];
                };
		struct  {
				  float4 row[4];
				};
       };

}matrix_internal;

class matrix
{
	private:

		inline void Mull(const matrix *m1, const matrix *m2)
		{
			 float tmp[16];

			 int i;
			 #define A(row,col)  m1->m.v[(col<<2)+row]
			 #define B(row,col)  m2->m.v[(col<<2)+row]
			 #define P(row,col)  tmp[(col<<2)+row]

			 for (i = 0; i < 4; i++)
			 {
		       float ai0=A(i,0),  ai1=A(i,1),  ai2=A(i,2),  ai3=A(i,3);
			   P(i,0) = ai0 * B(0,0) + ai1 * B(1,0) + ai2 * B(2,0) + ai3 * B(3,0);
			   P(i,1) = ai0 * B(0,1) + ai1 * B(1,1) + ai2 * B(2,1) + ai3 * B(3,1);
		       P(i,2) = ai0 * B(0,2) + ai1 * B(1,2) + ai2 * B(2,2) + ai3 * B(3,2);
			   P(i,3) = ai0 * B(0,3) + ai1 * B(1,3) + ai2 * B(2,3) + ai3 * B(3,3);
			 }

		     #undef A
		     #undef B
		     #undef P

			 for (i = 0; i < 16; i++)
	  	          m.v[i] = tmp[i];

			 return;
		}

    public:

        matrix_internal  m;

		matrix()
		{
			Identity();
		}

		void Identity()
		{
			for (int i = 0; i < 16; i++)
			     m.v[i] = 0.0f;
 
			m.xx = m.yy = m.zz = m.ww = 1.0f;

			return;
		}

		matrix operator * (const matrix &oth) const
		{
			matrix tmp;
			tmp.Mull((const matrix *)this, &oth);
			return tmp;
		}

		void operator *= (const matrix &oth)
		{
			matrix tmp;
			tmp.Mull((const matrix *)this, &oth);
			*this = tmp;
		}

		void MakeProjectionMatrix(float FOV, float Aspect, float NearPlane, float FarPlane);

		void MakeLookAtMatrix(float eyex,    float eyey,    float eyez,
		                      float centerx, float centery, float centerz,
					          float upx,     float upy,     float upz);

		void MakeRotation(float angle, float x, float y, float z)
		{
		   float mag, s, c;
		   float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;

		   s = sinf(angle);
		   c = cosf(angle);

		   mag = sqrtf( x*x + y*y + z*z );

		   if (mag < MATH_EPSILON)
		   {
			  this->Identity();
		      return;
		   }

		   x /= mag;
		   y /= mag;
		   z /= mag;

		   #define M(row,col)  m.v[(col<<2)+row]

		   xx = x * x;
		   yy = y * y;
		   zz = z * z;
		   xy = x * y;
		   yz = y * z;
		   zx = z * x;
		   xs = x * s;
		   ys = y * s;
		   zs = z * s;
		   one_c = 1.0f - c;

		   M(0,0) = (one_c * xx) + c;
		   M(0,1) = (one_c * xy) - zs;
		   M(0,2) = (one_c * zx) + ys;
		   M(0,3) = 0.0f;

		   M(1,0) = (one_c * xy) + zs;
		   M(1,1) = (one_c * yy) + c;
		   M(1,2) = (one_c * yz) - xs;
		   M(1,3) = 0.0f;

		   M(2,0) = (one_c * zx) - ys;
		   M(2,1) = (one_c * yz) + xs;
		   M(2,2) = (one_c * zz) + c;
		   M(2,3) = 0.0f;

		   M(3,0) = 0.0f;
		   M(3,1) = 0.0f;
		   M(3,2) = 0.0f;
		   M(3,3) = 1.0f;

		   #undef M
		   return;
		}
};
///////////////////////////////////////////////////////////////////////////////
#endif // __MATRIX_H_INCLUDED__
